import numpy as np
import pytest
import tensorflow as tf

from blacklight.engine.utils import data_utils


def test_split_data_has_one_batch_error():
    with pytest.raises(ValueError) as info:
        data_utils.split_dataset(
            tf.data.Dataset.from_tensor_slices(np.array([1, 2, 3])).batch(12),
            0.2,
        )

    assert "The dataset should at least contain 2 batches" in str(info.value)


def test_unzip_dataset_doesnt_unzip_single_dataset():
    dataset = tf.data.Dataset.from_tensor_slices(np.random.rand(10, 32, 2))
    dataset = data_utils.unzip_dataset(dataset)[0]
    dataset = data_utils.unzip_dataset(dataset)[0]
    assert data_utils.dataset_shape(dataset).as_list() == [32, 2]


def test_cast_to_string_with_float32():
    tensor = tf.constant([0.1, 0.2], dtype=tf.float32)
    assert tf.string == data_utils.cast_to_string(tensor).dtype


def test_cast_to_float32_from_float32():
    tensor = tf.constant([0.1, 0.2], dtype=tf.float32)
    assert tf.float32 == data_utils.cast_to_float32(tensor).dtype


def test_cast_to_float32_from_string():
    tensor = tf.constant(["0.3"], dtype=tf.string)
    assert tf.float32 == data_utils.cast_to_float32(tensor).dtype



from blacklight.genetic.base import Individual
from blacklight.genetic.base import Population
from blacklight.genetic.chromosomes import FeedForwardChromosome
from blacklight.engine import ModelConfig


class FeedForwardIndividual(Individual):
    """
    Individual for a feed forward neural network.

    Attributes:
        model_options (ModelConfig): ModelConfig Object containing options for the model (e.g. number of layers, number of neurons, etc.).
        problem_type (str): The problem type that the population is solving.
        num_classes (int): The number of classes for classification problems.
        model (Optional[Keras Model]): Keras model representing the individual's neural network.
        model_history (Optional[History]): Keras model training history.
        train_data (Optional[BlacklightDataset]): Training dataset.
        test_data (Optional[BlacklightDataset]): Testing dataset.
        chromosome (FeedForwardChromosome): Chromosome representing the individual's neural network topology.

    Args:
        options (ModelConfig): ModelConfig Object containing options for the model (e.g. number of layers, number of neurons, etc.).
        population (Population): The population object that this individual belongs to.
        parents_genes (Optional[Tuple[FeedForwardChromosome, FeedForwardChromosome]]): List of genes from parents.
    Examples:
        >>> options = ModelConfig(...)
        >>> population = Population(...)
        >>> individual1 = FeedForwardIndividual(options, population)
        >>> individual2 = FeedForwardIndividual(options, population)

        >>> fitness1 = individual1.get_fitness()
        >>> fitness2 = individual2.get_fitness()

        >>> child = individual1.mate(individual2)
        >>> child_fitness = child.get_fitness()
    """

    def __init__(
            self,
            options: ModelConfig,
            population: Population,
            parents_genes=None,
    ):
        """
        Initializes a FeedForwardIndividual.
        """
        super().__init__(parents_genes, population)
        self.model_options = options
        self.problem_type = population.problem_type
        self.num_classes = population.num_classes
        # Class parameters

        self.model = None
        self.model_history = None

        # data
        self.train_data = population.get_training_data()
        self.test_data = population.get_testing_data()

        # Create chromosome if parents genes are not provided
        if self.need_new_genes:
            self.chromosome = FeedForwardChromosome(
                self.model_options,
                genes=None,
                mutation_prob=0.1,
            )
        # Otherwise, create chromosome from parents genes
        else:
            self.chromosome = self._crossover(self.parents_genes)

    def _crossover(self, parents_chromosomes):
        """
        Perform crossover between the chromosomes of two parents to create a new chromosome.

        Args:
            parents_chromosomes (Tuple[FeedForwardChromosome, FeedForwardChromosome]): A tuple of parent chromosomes.

        Returns:
            FeedForwardChromosome: A new chromosome created by crossing over the parent chromosomes.

        """
        parent_one_chrome, parent_two_chrome = parents_chromosomes
        return FeedForwardChromosome.cross_over(
            parent_one_chrome, parent_two_chrome)

    def get_fitness(self):
        """
        Get the fitness of the individual.

        Returns:
            float: The fitness of the individual.
        """
        # Evaluate the model
        self.model = self.chromosome.get_model()
        fitness = self.chromosome.evaluate_model(
            self.train_data, self.test_data)
        self.fitness = fitness
        self.model_history = self.chromosome.get_model_history()

        return fitness

    def mate(self, other_individual):
        """
        Mate with another individual to produce a child.

        Args:
            other_individual (FeedForwardIndividual): The individual to mate with.

        Returns:
            FeedForwardIndividual: A new child individual created from the parents.
        """
        # Get genes from parents
        parents_genes = (self.chromosome, other_individual.chromosome)
        # Create child
        child = FeedForwardIndividual(
            parents_genes=parents_genes,
            population=self.population,
            options=self.model_options)
        # Return child
        return child



import pytest
from unittest.mock import MagicMock
from blacklight.genetic.chromosomes import FeedForwardChromosome
from blacklight.genetic.individuals import FeedForwardIndividual
from blacklight.engine import ModelConfig


@pytest.fixture
def model_config():
    model_options = {
        "problem_type": "binary_classification",
        "input_shape": 2,
        "min_dense_layers": 1,
        "max_dense_layers": 8,
        "min_dense_neurons": 1,
        "max_dense_neurons": 4,
        "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"],
        "num_classes": 3,
    }
    model_config = ModelConfig(model_options)
    return model_config


@pytest.fixture
def mock_population(model_config):
    population = MagicMock()
    population.problem_type = "classification"
    population.num_classes = 10
    population.get_training_data.return_value = (None, None)
    population.get_testing_data.return_value = (None, None)
    return population


def test_initialization(model_config, mock_population):
    individual = FeedForwardIndividual(model_config, mock_population)
    assert isinstance(individual, FeedForwardIndividual)
    assert isinstance(individual.chromosome, FeedForwardChromosome)
    assert individual.chromosome.model_params == model_config
    assert individual.chromosome is not None


def test_get_fitness(model_config, mock_population):
    individual = FeedForwardIndividual(model_config, mock_population)
    individual.chromosome.evaluate_model = MagicMock(return_value=0.75)
    fitness = individual.get_fitness()
    assert fitness == 0.75


def test_mate(model_config, mock_population):
    individual1 = FeedForwardIndividual(model_config, mock_population)
    individual2 = FeedForwardIndividual(model_config, mock_population)
    child = individual1.mate(individual2)
    assert isinstance(child, FeedForwardIndividual)
    assert child.chromosome != individual1.chromosome
    assert child.chromosome != individual2.chromosome



from __future__ import annotations
from blacklight.genetic.base import Chromosome
from blacklight.engine import BlacklightModel
from blacklight.engine import ModelConfig
from typing import List, Optional, Tuple
from keras import Sequential

import random


class FeedForwardChromosome(Chromosome):
    """
    Feed forward chromosome class representing a feed forward neural network.

    Args:
        model_params ModelConfig: The parameters for the Keras model. Defaults to None.
        genes (Optional[List[Tuple[int, str]]], optional): The list of genes representing the neural network structure. Defaults to None.
        mutation_prob (Optional[int], optional): The probability of mutation for the chromosome. Defaults to None.

    Examples:
        >>> chromosome = FeedForwardChromosome(input_shape=10)
    """

    def __init__(self,
                 model_params: ModelConfig,
                 genes: Optional[List[Tuple[int, str]]] = None,
                 mutation_prob: Optional[float] = None,
                 ):
        super().__init__()

        # Get the model params to build chromosome
        self.model_params = model_params

        # Check if the chromosome has new genes from parents
        has_new_genes = genes is not None

        # Get the mutation probability
        self.mutation_prob = mutation_prob

        # Get random genes if no genes are provided from parents.
        self.genes = genes if genes else self._random_genes()

        # Set the length of the chromosome
        self.length = len(self.genes)

        # If we got genes from parents, we need to mutate them.
        if has_new_genes:
            self._mutate()

        # Create the Keras model based on the genes, and the model params.
        self.model = BlacklightModel(
            self.model_params, self.genes)
        self.model.create_model()

    def _random_genes(self) -> List[Tuple[str, int, str]]:
        """
        Generate random feed forward genes for the chromosome.

        Returns:
            List[Tuple[int, str]]: A list of genes representing the neural network structure.
        """
        layer_size = range(
            self.model_params.get("min_dense_neurons"),
            self.model_params.get("max_dense_neurons"))

        genes = [
            ("Dense",
             random.choice(layer_size),
             random.choice(
                 self.model_params.get("dense_activation_types"))) for _ in range(
                random.choice(
                    range(
                        self.model_params.get("min_dense_layers"),
                        self.model_params.get("max_dense_layers"))))]

        return genes

    @staticmethod
    def cross_over(chromosome_a: FeedForwardChromosome,
                   chromosome_b: FeedForwardChromosome) -> FeedForwardChromosome:
        """
        Handle feed forward cross over between two chromosomes.

        Args:
            chromosome_a (FeedForwardChromosome): The first chromosome to perform cross over.
            chromosome_b (FeedForwardChromosome): The second chromosome to perform cross over.

        Returns:
            FeedForwardChromosome: A new chromosome created by cross over between the input chromosomes.

        Examples:
            >>> chromosome_a = FeedForwardChromosome(input_shape=10)
            >>> chromosome_b = FeedForwardChromosome(input_shape=10)
            >>> new_chromosome = FeedForwardChromosome.cross_over(chromosome_a,chromosome_b)
        """
        shorter_chromosome, longer_chromosome = Chromosome.get_shortest_chromosome(
            chromosome_a, chromosome_b)

        points = random.randint(1, len(shorter_chromosome.genes))
        base_one = shorter_chromosome.genes[:points]
        link_one = longer_chromosome.genes[points:]

        base_two = longer_chromosome.genes[:points]
        link_two = shorter_chromosome.genes[points:]

        recombinant_one = base_one + link_one
        recombinant_two = base_two + link_two

        genes = random.choice([recombinant_one, recombinant_two])

        new_chromosome = FeedForwardChromosome(
            model_params=chromosome_a.model_params,
            genes=genes,
            mutation_prob=chromosome_a.mutation_prob,
        )

        return new_chromosome

    def _mutate(self) -> None:
        """
        Mutate the chromosome by randomly changing one of its genes.
        """
        mutate = random.choices([True, False], weights=[
            self.mutation_prob, 1 - self.mutation_prob], k=1)[0]

        if mutate:
            layer_idx = random.choice(range(self.length))
            new_allele = random.randint(
                1, self.model_params.get("max_dense_neurons"))
            new_activation = random.choice(
                self.model_params.get("dense_activation_types"))
            self.genes[layer_idx] = ("Dense", new_allele, new_activation)

    def evaluate_model(self, train_data, test_data):
        """
        Evaluate the model and return the loss and accuracy.

        Returns:
            Tuple[float, float]: The loss and accuracy of the model.
        """
        fitness = self.model.evaluate_model(train_data, test_data)
        return fitness

    def get_model(self) -> Sequential:
        """
        Get the Keras Sequential model of the chromosome.

        Returns:
            tf.keras.Sequential: The feed forward neural network model.

        Examples:
            >>> chromosome = FeedForwardChromosome(input_shape=10)
            >>> model = chromosome.get_model()
        """
        return self.model.get_model()

    def get_model_history(self):
        return self.model.get_model_history()

    def __repr__(self):
        return f"FeedForwardChromosome with genes: {self.genes}"

    def __str__(self):
        return f"FeedForwardChromosome with genes: {self.genes} \n and model: {self.model.summary()}"



from abc import ABC, abstractmethod
from typing import Tuple


class Chromosome(ABC):
    """
    Abstract class for a chromosome in an individual. Implements common traits between all chromosomes.
    """

    def __init__(self):
        self.genes = None
        self.length = None

    @abstractmethod
    def _random_genes(self):
        """
        Randomly initialize genes for each type of chromosome if there are no parents.
        """
        pass

    @abstractmethod
    def cross_over(self, *args):
        """
        Handle cross over between two chromosomes.
        """
        pass

    @abstractmethod
    def _mutate(self):
        """
        Mutate genes of this chromosome.
        """
        pass

    @staticmethod
    def get_shortest_chromosome(
            chromosome_a: object, chromosome_b: object) -> Tuple[object, object]:
        """
        Determine the shortest and longest chromosomes.

        Args:
            chromosome_a (BaseChromosome): The first chromosome to compare.
            chromosome_b (BaseChromosome): The second chromosome to compare.

        Returns:
            Tuple[BaseChromosome, BaseChromosome]: A tuple containing the shortest chromosome and the longest chromosome, in that order.
        """
        if chromosome_a.length > chromosome_b.length:
            return chromosome_b, chromosome_a
        else:
            return chromosome_a, chromosome_b



import numpy as np
import tensorflow as tf
import pandas as pd
import math
from sklearn.preprocessing import LabelEncoder
from keras.utils import to_categorical


def parse_target_column(data):
    """
    Parse target column from data

    Parameters:
        - data: The input data to be classified. Must be of type pandas.DataFrame.

    Returns:
        - X: The feature columns of the input data.
        - y: The target column of the input data.
    """
    possible_prediction_columns = ["label", "prediction", "labels"]
    y_column_name = ""
    for column in data.columns:
        if str(column).lower() in possible_prediction_columns:
            y_column_name = column
            break

    y = data[y_column_name]
    X = data.to_numpy()[:, :-1]
    return X, y


def determine_input_data_type(X):
    """
    Determine the type of input data.
    Current supported datatypes are numpy.ndarray, pandas.DataFrame, tf.data.Dataset, a file location, or a directory.

    Parameters:
        - X: The input data to be classified. Must be of type numpy.ndarray, pandas.DataFrame, tf.data.Dataset, a file
             location, or a directory.
    """
    if isinstance(X, pd.DataFrame):
        return "pandas"
    elif isinstance(X, np.ndarray):
        return "numpy"
    elif isinstance(X, tf.data.Dataset):
        return "tf.data"
    elif isinstance(X, str):
        if "." in X:
            return "file"
        else:
            return "directory"
    else:
        raise TypeError(
            "X must be of type numpy.ndarray, pandas.DataFrame, tf.data.Dataset, a file location, "
            "or a directory.")


def handle_pandas_data(X, y):
    """
    Handle pandas data by parsing the target column and encoding the target column if it is a string.

    Parameters:
        - X: The input data to be classified. Must be of type pandas.DataFrame.
        - y: The target column of the input data. Must be of type pandas.Series.
    """
    try:
        if y is None:
            X, y = parse_target_column(X)
        return np.array(X).astype('float32'), np.array(y)
    except BaseException:
        raise ValueError(
            "X (dataframe) could not be separated into label and feature columns.")


def handle_numpy_data(X, y):
    try:
        if y is None:
            X, y = X[:, :-1], X[:, -1]
        return X, y
    except IndexError:
        raise ValueError(
            "X (np.array) could not be separated into label and feature columns.")


def handle_tf_data(X, y):
    return None, None


def handle_file_data(X):
    df = read_data_from_file(X, X.split(".")[-1])
    X, y = handle_pandas_data(df, None)
    return X, y


class BlacklightDataset(tf.keras.utils.Sequence):
    """
    A dataset for all blacklight populations to use. This class is a wrapper for tf.keras.utils.Sequence
    that handles all data types and formats, which also allows for multiprocessing during training.
    This class is used by the Individual class to extract data in the fit method for processing.

    Parameters:
        - X: The data to be used for training. This can be a numpy array, pandas dataframe, or a file location.
        - y: The labels for the data. This can be a numpy array, pandas series. If not provided, the last column of X will be used or the column labeled "label" or "labels".
        - batch_size: The size of the batch to be used for training. If not provided, the batch size will be the length of the data.
    """

    def __init__(self, X, y=None, batch_size=None):

        self.X = X
        self.y = y
        if y is None:
            type_of_data = determine_input_data_type(X)
            if type_of_data == "pandas":
                self.X, self.y = handle_pandas_data(self.X, self.y)
            elif type_of_data == "numpy":
                self.X, self.y = handle_numpy_data(self.X, self.y)
            elif type_of_data == "tf.data":
                self.X, self.y = handle_tf_data(self.X, self.y)
            elif type_of_data == "file":
                self.X, self.y = handle_file_data(self.X)
        self.batch_size = batch_size if batch_size else len(self.X)
        self.X_shape = self.X.shape[-1]
        # If the type of the target is string, encode it
        if isinstance(self.y[0], str):
            self.one_hot_encode_target()

    def one_hot_encode_target(self):
        y = LabelEncoder().fit_transform(self.y)
        self.y = to_categorical(y)

    def __len__(self):
        return math.ceil(len(self.X) / self.batch_size)

    def __getitem__(self, idx):
        """
        Grab a batch of data for use in training. If batch_size is not provided, the batch size will be the length of the data.

        Parameters:
            - idx: The index of the batch to be returned.
        """
        batch_x = self.X[idx * self.batch_size:(idx + 1) *
                         self.batch_size]
        batch_y = self.y[idx * self.batch_size:(idx + 1) *
                         self.batch_size]
        return batch_x, batch_y


def read_data_from_file(data_location, data_type, **kwargs) -> pd.DataFrame:
    """
    Read data from a file location. This function will attempt to read the data as a CSV, JSON, Excel, or Parquet file.

    Parameters:
        - data_location: The location of the data to be read.
        - data_type: The type of data to be read. Must be one of 'csv', 'json', 'excel', or 'parquet'.
        - **kwargs: Keyword arguments to be passed to the read function.

    Returns:
        - read_data: The data read from the file location.
    """
    try:
        # Attempt to read CSV Data
        if data_type == "csv":
            read_data = pd.read_csv(data_location, **kwargs)
        # Attempt to read JSON Data
        elif data_type == "json":
            read_data = pd.read_json(data_location, **kwargs)
        # Attempt to read Excel Data
        elif data_type == "excel":
            read_data = pd.read_excel(data_location, **kwargs)
        # Attempt to read Parquet Data
        elif data_type == "parquet":
            read_data = pd.read_parquet(data_location, **kwargs)
        # Raise ValueError if data_type is not one of the above
        else:
            raise ValueError(
                "data_type must be one of 'csv', 'json', 'excel', or 'parquet'")
    # Raise FileNotFoundError if data_location is not found in any of above
    # file types
    except FileNotFoundError:
        raise FileNotFoundError(f"{data_type} file not found at data_location")

    return pd.DataFrame(read_data)



import tensorflow as tf
from blacklight.base.individual import Individual
from blacklight.autoML.individuals.individualutils import FeedForwardConstants
from collections import OrderedDict


class ConvolutionalIndividual(Individual):
    """
    Individual for a convolutional neural network.

    Parameters:
            parents_genes: list of genes from parents
            population: population object that this individual belongs to
            MAX_NEURONS: maximum number of neurons allowed
            MAX_LAYERS: maximum number of layers allowed
            MAX_
            NUM_PARENTS: number of parents used for crossover
            **kwargs: arguments to be passed to Keras

    Returns:
        None
    """

    def __init__(
            self,
            parents_genes=None,
            population=None,
            MAX_NEURONS=64,
            MAX_LAYERS=10,
            NUM_PARENTS=2,
            **kwargs):
        """
        Initializes a FeedForwardIndividual.
        """
        super().__init__(parents_genes, population)
        # Feed Forward Neural Network parameters
        self.MAX_LAYERS = MAX_LAYERS
        self.MAX_NEURONS = MAX_NEURONS
        self.NUM_PARENTS = NUM_PARENTS
        # Keras parameters -> Maybe these need to be in the population, not the
        # individual?
        self.EPOCHS = kwargs.get("EPOCHS", 1000)
        self.BATCH_SIZE = kwargs.get("BATCH_SIZE", 2048)
        self.VERBOSE = kwargs.get("VERBOSE", 0)
        self.OPTIMIZER = kwargs.get("OPTIMIZER", tf.keras.optimizers.Adam)
        self.VALIDATION_SPLIT = kwargs.get("VALIDATION_SPLIT", None)
        self.OUTPUT_BIAS = kwargs.get("OUTPUT_BIAS", True)
        self.LOSS = kwargs.get("LOSS", 'categorical_crossentropy')
        self.METRICS = kwargs.get(
            "METRICS", FeedForwardConstants.default_metrics)
        self.LEARNING_RATE = kwargs.get("LEARNING_RATE", .0001)
        # Training parameters
        self.EARLY_STOPPING = kwargs.get("EARLY_STOPPING", True)
        self.CALLBACKS = kwargs.get(
            "CALLBACKS", FeedForwardConstants.default_early_stopping)
        # Class parameters
        self.dominant_gene = None
        self.genes = self._random_genes() if self.need_new_genes else self._crossover()

    def _random_genes(self):
        """
        Randomly generate genes for this individual.
        """
        genes = OrderedDict()
        for layer in range(self.MAX_LAYERS):
            genes[layer] = OrderedDict()



import random
import pytest
from blacklight.genetic.chromosomes import FeedForwardChromosome
from blacklight.engine import ModelConfig
from unittest.mock import MagicMock


def get_model_config() -> ModelConfig:
    model_options = {
        "problem_type": "binary_classification",
        "input_shape": 10,
        "min_dense_layers": 1,
        "max_dense_layers": 10,
        "min_dense_neurons": 5,
        "max_dense_neurons": 10,
        "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"],
        "num_classes": 3,
    }
    model_config = ModelConfig(model_options)
    return model_config


@pytest.fixture
def seed():
    random.seed(42)


@pytest.mark.usefixtures("seed")
def test_feed_forward_chromosome_init(seed):
    model_config = get_model_config()
    ff_chromosome = FeedForwardChromosome(model_params=model_config)

    expected_genes = [("Dense", 5, "tanh"), ("Dense", 6, "sigmoid")]
    assert ff_chromosome.genes == expected_genes


def test_feed_forward_chromosome_crossover(seed):
    model_config = get_model_config()

    # [(Dense, 5, tanh), (Dense, 6, sigmoid)]
    chromosome_a = FeedForwardChromosome(
        model_params=model_config, mutation_prob=0.01)

    # [(Dense, 5, relu), (Dense, 9, selu), (Dense, 5, relu)]
    chromosome_b = FeedForwardChromosome(
        model_params=model_config, mutation_prob=0.01)

    new_chromosome = FeedForwardChromosome.cross_over(
        chromosome_a, chromosome_b)

    assert new_chromosome.genes is not None


def test_feed_forward_chromosome_mutate(seed):
    model_config = get_model_config()
    ff_chromosome = FeedForwardChromosome(
        model_params=model_config, mutation_prob=0.99)
    ff_chromosome._mutate()

    assert ff_chromosome.genes is not None


def test_feed_forward_chromosome_evaluate_model(seed):
    model_config = get_model_config()
    ff_chromosome = FeedForwardChromosome(model_params=model_config)

    # Mock the evaluate_model method on the BlacklightModel class
    ff_chromosome.model.evaluate_model = MagicMock(return_value=0.8)

    train_data = [100, 10]
    test_data = [20, 10]

    fitness = ff_chromosome.evaluate_model(train_data, test_data)
    assert fitness == 0.8



"""Converts the Iris.csv file to other datatypes for testing"""
# Imports
import pandas as pd

# Read the df
df = pd.read_csv("test/dataloaders/data/Iris.csv")
# Json
df.to_json("test/dataloaders/data/Iris.json")
# Parquet
df.to_parquet("test/dataloaders/data/Iris.parquet")
# Excel
df.to_excel("test/dataloaders/data/Iris.xlsx", index=False)



import tensorflow as tf


class BlacklightModel:
    def __init__(self, model_config, genes):
        self.model_history = None
        self.model_config = model_config
        self.genes = genes
        self.model = None

    def create_model(self):
        model = tf.keras.Sequential()
        model.add(
            tf.keras.layers.InputLayer(
                input_shape=self.model_config.get("input_shape")))
        for gene in self.genes:
            if gene[0] == "Conv2D":
                model.add(
                    tf.keras.layers.Conv2D(
                        gene[1],
                        gene[2],
                        activation=gene[3]))
            elif gene[0] == "MaxPooling2D":
                model.add(tf.keras.layers.MaxPooling2D(gene[1]))
            elif gene[0] == "Flatten":
                model.add(tf.keras.layers.Flatten())
            elif gene[0] == "Dense":
                model.add(tf.keras.layers.Dense(gene[1], activation=gene[2]))
            else:
                raise ValueError(f"Invalid gene type: {gene[0]}")

        target_layer = self.model_config.get("target_layer")
        model.add(
            tf.keras.layers.Dense(
                target_layer[0],
                activation=target_layer[1]))

        model.compile(
            optimizer=self.model_config.get("optimizer"),
            loss=self.model_config.get("loss"),
            metrics=self.model_config.get("metrics")
        )
        self.model = model

    def train_model(self, train_data):
        if self.model:
            self.model_history = self.model.fit(
                x=train_data,
                epochs=self.model_config.get("epochs"),
                batch_size=self.model_config.get("batch_size"),
                validation_split=self.model_config.get("validation_split"),
                verbose=self.model_config.get("verbose"),
                class_weight=self.model_config.get("class_weight"),
                validation_data=self.model_config.get("validation_data"),
                callbacks=self.model_config.get("callbacks") if self.model_config.get("early_stopping") else None)
        else:
            raise ValueError("Model not created yet. Please call create_model() first.")

    def get_model(self):
        return self.model

    def get_model_history(self):
        return self.model_history

    def evaluate_model(self, train_data, test_data):
        # first train the model
        if self.model_history is None:
            self.train_model(train_data)
        # then evaluate the model
        if self.model:
            results = self.model.evaluate(
                test_data,
                batch_size=self.model_config.get("batch_size"),
                verbose=self.model_config.get("verbose"),
                return_dict=True
            )
        else:
            raise ValueError("Model not created yet. Please call create_model() first.")
        # the fitness metric is extracted.
        fitness = results[self.model_config.get("fitness_metric")]
        return fitness



import tensorflow as tf
from unittest.mock import MagicMock
from blacklight.engine import BlacklightModel
from blacklight.engine import ModelConfig

# Helper function to create a ModelConfig instance


def get_model_config():
    config = {
        "target_layer": (1, "sigmoid"),
        "loss": "binary_crossentropy",
        "optimizer": "adam",
        "metrics": ModelConfig.get_default_metrics(),
        "learning_rate": 0.001,
        "epochs": 1000,
        "batch_size": 32,
        "verbose": 0,
        "validation_split": 0.1,
        "early_stopping": True,
        "callbacks": ModelConfig.get_default_callbacks(),
        "input_shape": (32, 32, 3),
        "fitness_metric": "auc"
    }
    return ModelConfig(config)


# Helper function to create a sample gene list
def get_sample_genes():
    return [
        ("Conv2D", 32, (3, 3), "relu"),
        ("MaxPooling2D", (2, 2)),
        ("Flatten",),
        ("Dense", 128, "relu"),
    ]


def test_blacklight_model_init():
    model_config = get_model_config()
    genes = get_sample_genes()
    bl_model = BlacklightModel(model_config, genes)
    assert bl_model.model_config == model_config
    assert bl_model.genes == genes
    assert bl_model.model is None


def test_blacklight_model_create_model():
    model_config = get_model_config()
    genes = get_sample_genes()
    bl_model = BlacklightModel(model_config, genes)
    bl_model.create_model()
    assert isinstance(bl_model.model, tf.keras.Model)


def test_blacklight_model_train_model():
    model_config = get_model_config()
    genes = get_sample_genes()
    bl_model = BlacklightModel(model_config, genes)
    bl_model.create_model()
    # Mock the fit method to avoid actual training
    bl_model.model.fit = MagicMock(return_value="Some Training History")
    bl_model.train_model(tf.random.normal([100, 32, 32, 3]))
    assert bl_model.model_history is not None


def test_blacklight_model_evaluate_model():
    model_config = get_model_config()
    genes = get_sample_genes()
    bl_model = BlacklightModel(model_config, genes)
    bl_model.create_model()
    # Mock the fit and evaluate methods to avoid actual training and evaluation
    bl_model.model.fit = MagicMock(return_value=None)
    bl_model.model.evaluate = MagicMock(return_value={"auc": 0.8})
    fitness = bl_model.evaluate_model(tf.random.normal(
        [100, 32, 32, 3]), tf.random.normal([20, 32, 32, 3]))
    assert fitness == 0.8



from blacklight.genetic.individuals import FeedForwardIndividual
from blacklight.engine.data import BlacklightDataset
from blacklight.genetic.base import Population
from blacklight.engine import ModelConfig
from collections import OrderedDict
import matplotlib.pyplot as plt
from tqdm import tqdm
from sklearn.model_selection import train_test_split
import logging


class FeedForward(Population):
    """
    A population of Feed Forward DNN topologies that will be evaluated utilizing a genetic search.

    Parameters:
            number_of_individuals (int): The number of individuals that the user wants in the population.
            num_parents_mating (int): The number of parents that will be used to mate and create the next generation.
            death_percentage (float): The percentage of individuals that will die off each generation.
            number_of_generations (int): The number of generations that the population will be simulated for.
            options (ModelConfig): ModelConfig Object containing options for the model (e.g. number of layers, number of neurons, etc.).
    """

    def __init__(
        self,
        number_of_individuals,
        num_parents_mating,
        death_percentage,
        number_of_generations,
        options,
    ):
        super().__init__(
            number_of_individuals,
            num_parents_mating,
            death_percentage,
            number_of_generations,
            options,
        )
        self.num_classes = None
        self.test_data = None
        self.problem_type_individual = None
        self.model_history = None
        self.model = None
        self.data = None
        self.num_individuals = number_of_individuals
        self.num_parents_mating = num_parents_mating
        self.num_generations = number_of_generations
        self.death_percentage = death_percentage
        self.problem_type = None
        self.options = ModelConfig(options)

    def _initialize_individuals(self):
        """
        Initialize the individuals in this population.
        Returns:

        """
        self.individuals = OrderedDict(
            {
                FeedForwardIndividual(self.options, self, None): f"{i}"
                for i in range(self.num_individuals)
            }
        )

    def fit(self, X_train, y_train=None, X_test=None, y_test=None, **kwargs):
        """
        Fit this population of FeedForward Individuals to the given data, and return the best model.
        You can pass ONLY x if and only if you have a column in your dataset (if pandas) called labels or label.
        You can also do this if you have a numpy array with the label column at position -1.
        Otherwise, you must pass X_train and y_train. If you pass X_test and y_test, then the model will be evaluated
        on the test data after each generation. If not, then test train split will be applied with an 80/20 split.

        Parameters:
                X: The data to fit this population to. Has to be either a pandas dataframe with columns: [feature1, feature2, ..., featureN, label] or a file path to a file of formats specified in `ref: dataLoaders.dataLoader.choose_data_loader` that have the same layout.
                **kwargs: Any additional arguments to pass to each :class:`~blacklight.autoML.individuals.FeedForwardIndividual` which is a Keras model.
                y: The labels of the data. If None, then the labels are assumed to be the last column of the data.
        """
        # Disable tensorflow logging
        import os

        logging.getLogger("tensorflow").disabled = True
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

        if X_test is not None and y_test is not None:
            self.test_data = BlacklightDataset(
                X_test, y_test, kwargs.get("batch_size", None)
            )
            self.train_data = BlacklightDataset(
                X_train, y_train, kwargs.get("batch_size", None)
            )
        else:
            data = BlacklightDataset(X_train, y_train, kwargs.get("batch_size", None))
            X_train, X_test, y_train, y_test = train_test_split(
                data.X, data.y, test_size=0.2
            )
            self.test_data = BlacklightDataset(
                X_test, y_test, kwargs.get("batch_size", None)
            )
            self.train_data = BlacklightDataset(
                X_train, y_train, kwargs.get("batch_size", None)
            )

        # Initialize individuals
        self._initialize_individuals()

        # Simulate the population for the specified number of generations
        self._simulate()

        # Get the best individual
        self.best_individual = list(self.individuals.keys())[0]
        self.model, self.model_history = (
            self.best_individual.model,
            self.best_individual.model_history,
        )

    def print_model_summary(self):
        """
        Print the summary of the best model.
        """
        self.model.summary()

    def print_model_training_history(self):
        plt.plot(self.model_history.history["accuracy"])
        plt.title("model accuracy")
        plt.ylabel("accuracy")
        plt.xlabel("epoch")
        plt.show()

    def predict(self, X):
        """
        Predict the labels of the given data, utilizing the best found model.

        Parameters:
                X: The data to predict the labels of. Has to be either a pandas dataframe with columns: [feature1, feature2, ..., featureN] or a file path to a file of formats specified in `ref: dataLoaders.dataLoader.choose_data_loader` that have the same layout.
        """
        return self.model.predict(X)

    def fit_predict(self, data, **kwargs):
        """
        Fit this population of FeedForward Individuals to the given data, and return the predictions of the best model.

        Parameters:
                data: The data to fit this population to. Has to be either a pandas dataframe with columns: [feature1, feature2, ..., featureN, label] or a file path to a file of formats specified in `ref: dataLoaders.dataLoader.choose_data_loader` that have the same layout.
        """
        self.fit(data, **kwargs)
        return self.predict(self.get_testing_data())

    def _simulate(self):
        """
        Simulate the population for the number of generations.
        Callable method.
        :return:
        """
        print(
            f"\nSimulating feed forward neural network population with {self.num_individuals} individuals for {self.num_generations} generations."
        )
        for gen in range(self.num_generations):
            print(f"\nGeneration {gen}")
            print(f"Evaluating Individuals in generation {gen}")
            self._evaluate()
            print(f"\nReproducing Individuals in generation {gen}")
            self._reproduce()

    def _evaluate(self):
        """
        Evaluate the fitness of all individuals in the population.
        :return:
        """
        evaluated_individuals = OrderedDict(
            {
                individual: individual.get_fitness()
                for individual in tqdm(self.individuals.keys())
            }
        )
        self.individuals = OrderedDict(
            sorted(evaluated_individuals.items(), key=lambda x: x[1], reverse=True)
        )

    def _get_fitness(self):
        pass



from __future__ import annotations
import random
from abc import ABC
import tensorflow as tf
from blacklight.genetic.base.chromosome import Chromosome
from blacklight.engine import BlacklightModel


class ConvolutionalChromosome(Chromosome, ABC):
    """
    Convolutional chromosome class, which represents a convolutional neural network architecture.

    The class inherits from BaseChromosome and implements methods to generate random genes, mutate
    the chromosome, and create a Keras model based on the genes. The genes represent both the
    convolutional and dense layers of the neural network.

    Parameters:
        input_shape (tuple): The input shape of the model (width, height, channels).
        max_conv_layers (int): The maximum number of convolutional layers in the model.
        max_dense_layers (int): The maximum number of dense layers in the model.
        genes (list): List of genes from parents (default: None).
        mutation_prob (float): Probability of mutation (default: None).
        model_params (dict): Dictionary of model parameters such as optimizer, loss, metrics, and learning rate.

    Attributes:
        model (tf.keras.Model): The Keras model represented by the chromosome.
        genes (list): The list of genes that describe the architecture of the neural network.
        length (int): The number of genes in the chromosome.

    Raises:
        ValueError: If the input_shape is not a tuple or if the mutation probability is not between 0 and 1.

    Methods:
        _random_genes: Generate random genes for the chromosome.
        _make_model: Create a Keras model based on the chromosome's genes.
        _mutate: Mutate the chromosome's genes.
        get_model: Return the Keras model.

    Examples:
        >>> input_shape = (32, 32, 3)
        >>> max_conv_layers = 5
        >>> max_dense_layers = 3
        >>> model_params = {
            "optimizer": tf.keras.optimizers.Adam,
            "loss": 'categorical_crossentropy',
            "metrics": ['accuracy'],
            "learning_rate": 0.001
        }
        >>> chromosome = ConvolutionalChromosome(input_shape, max_conv_layers, max_dense_layers, model_params=model_params)
        >>> model = chromosome.get_model()
    """

    def __init__(
            self,
            input_shape=None,
            max_conv_layers=10,
            max_dense_layers=5,
            min_conv_layers=1,
            min_dense_layers=1,
            genes=None,
            mutation_prob=None,
            model_params=None):
        super().__init__()
        self.model_params = model_params if model_params else {}
        has_new_genes = genes is not None

        self.length = None
        self.mutation_prob = mutation_prob

        self.genes = genes if genes else self._random_genes()
        self.length = len(self.genes)
        if has_new_genes:
            self._mutate()

        model = BlacklightModel(self.model_params, self.genes)

        self.model = model.create_model()

    @staticmethod
    def split_convolutional_dense(genes: list) -> tuple[list, list]:
        """
        Split convolutional and dense layers from the genes.

        Parameters:
            genes: list of alleles for a convolutional chromosome

        Returns:
            - conv_part: the convolutional part of the genes
            - dense_part: the dense part of the genes
        """
        try:
            flatten_index = genes.index(("Flatten",))
        except ValueError:
            raise ValueError("Flatten layer not found in the genes.")

        conv_part = genes[:flatten_index]
        dense_part = genes[flatten_index + 1:]
        return conv_part, dense_part

    @staticmethod
    def cross_over(
            chromosome_a: ConvolutionalChromosome,
            chromosome_b: ConvolutionalChromosome) -> ConvolutionalChromosome:
        """
            Perform a crossover operation between two ConvolutionalChromosome instances.

            This function creates two new offspring by crossing over the genes of the input chromosomes.
            The crossover operation is applied separately to the convolutional part and the dense part
            of the chromosomes, respecting their structure.

            Args:
                chromosome_a (ConvolutionalChromosome): The first parent chromosome.
                chromosome_b (ConvolutionalChromosome): The second parent chromosome.

            Returns:
                ConvolutionalChromosome: A new ConvolutionalChromosome instances created through crossover.

            Raises:
                ValueError: If the "flatten" layer is not found in the genes or if the input chromosomes have different input shapes.
            """
        shorter_chromosome, longer_chromosome = Chromosome.get_shortest_chromosome(
            chromosome_a, chromosome_b)

        shorter_conv_part, shorter_dense_part = ConvolutionalChromosome.split_convolutional_dense(
            shorter_chromosome.genes)
        longer_conv_part, longer_dense_part = ConvolutionalChromosome.split_convolutional_dense(
            longer_chromosome.genes)

        crossover_point_conv = random.randint(1, len(shorter_conv_part))
        crossover_point_dense = random.randint(1, len(shorter_dense_part))

        recombinant_conv_part = shorter_conv_part[:crossover_point_conv] + \
            longer_conv_part[crossover_point_conv:]
        recombinant_dense_part = shorter_dense_part[:crossover_point_dense] + \
            longer_dense_part[crossover_point_dense:]

        recombinant_genes = recombinant_conv_part + \
            [("Flatten",)] + recombinant_dense_part

        new_chromosome = ConvolutionalChromosome(
            input_shape=chromosome_a.input_shape,
            mutation_prob=chromosome_a.mutation_prob,
            genes=recombinant_genes,
            model_params=chromosome_a.model_params)

        return new_chromosome

    def _mutate(self) -> None:
        """
        Mutate the chromosome's genes while preserving the structure.

        This method applies random mutations to the genes of the chromosome while ensuring that
        convolutional layers remain in the convolutional section and dense layers remain in the
        dense section. It selects a random layer for mutation and changes its parameters based on
        the layer type.

        Raises:
            ValueError: If an unexpected layer type is encountered in the genes.
        """
        mutate = random.choices([True, False], weights=[
            self.mutation_prob, 10 - self.mutation_prob], k=1)[0]

        if mutate:
            conv_indices = [
                i for i, gene in enumerate(
                    self.genes) if gene[0] == "Conv2D"]
            dense_indices = [
                i for i, gene in enumerate(
                    self.genes) if gene[0] == "Dense"]

            mutation_type = random.choice(["conv", "dense"])
            if mutation_type == "conv":
                layer_idx = random.choice(conv_indices)
                new_filters = random.randint(32, 256)
                self.genes[layer_idx] = (
                    "Conv2D",
                    new_filters,
                    self.genes[layer_idx][2],
                    self.genes[layer_idx][3])
            elif mutation_type == "dense":
                layer_idx = random.choice(dense_indices)
                new_neurons = random.randint(32, 256)
                self.genes[layer_idx] = (
                    "Dense", new_neurons, self.genes[layer_idx][2])

    def get_model(self) -> tf.keras.Model:
        return self.model

    def __repr__(self):
        return f"ConvolutionalChromosome with genes: {self.genes}"

    def __str__(self):
        return f"ConvolutionalChromosome with genes: {self.genes} \n and model: {self.model.summary()}"



from blacklight.genetic.populations import FeedForward
from blacklight.genetic.individuals import FeedForwardIndividual
from blacklight.engine import ModelConfig

import numpy as np
import pandas as pd

import unittest.mock
import pytest
from unittest.mock import MagicMock
from collections import OrderedDict
import random


@pytest.fixture
def getdata():
    X = np.array([[5.1, 3.5, 1.4, 0.2], [4.9, 3.0, 1.4, 0.2], [4.7, 3.2, 1.3, 0.2], [
        4.6, 3.1, 1.5, 0.2], [4.7, 3.2, 1.3, 0.2], [4.7, 3.2, 1.3, 0.2], [4.7, 3.2, 1.3, 0.2]])
    y = np.array(['Iris-setosa',
                  'Iris-versicolor',
                  'Iris-virginica',
                  'Iris-setosa',
                  'Iris-versicolor',
                  'Iris-virginica',
                  'Iris-setosa'])
    dataDF = pd.DataFrame(
        data=X,
        columns=[
            'SepalLengthCm',
            'SepalWidthCm',
            'PetalLengthCm',
            'PetalWidthCm'])
    dataDF['label'] = y

    return X, y


@pytest.fixture
def get_test_data():
    X = np.array([[5.1, 3.5, 1.4, 0.2], [4.9, 3.0, 1.4, 0.2], [
        4.6, 3.1, 1.5, 0.2], [4.7, 3.2, 1.3, 0.2], [4.7, 3.2, 1.3, 0.2]])
    y = np.array(['Iris-setosa',
                  'Iris-versicolor',
                  'Iris-virginica',
                  'Iris-setosa',
                  'Iris-setosa'])
    dataDF = pd.DataFrame(
        data=X,
        columns=[
            'SepalLengthCm',
            'SepalWidthCm',
            'PetalLengthCm',
            'PetalWidthCm'])
    dataDF['label'] = y

    return X, y


@pytest.fixture
def model_config():
    model_options = {
        "layer_information": {
            "problem_type": "classification",
            "input_shape": 4,
            "min_dense_layers": 1,
            "max_dense_layers": 8,
            "min_dense_neurons": 2,
            "max_dense_neurons": 8,
            "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"]
        },
        "problem_type": "classification",
        "num_classes": 3,
    }
    model_config = ModelConfig(model_options)
    return model_config.config


@pytest.fixture
def model_options():
    model_options = {
        "layer_information": {
            "problem_type": "classification",
            "input_shape": 4,
            "min_dense_layers": 1,
            "max_dense_layers": 8,
            "min_dense_neurons": 2,
            "max_dense_neurons": 8,
            "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"]
        },
        "problem_type": "classification",
        "num_classes": 3,
    }
    return model_options


def test_initialization(model_options):
    population = FeedForward(10, 5, 0.5, 5, model_options)
    assert isinstance(population, FeedForward)
    assert isinstance(population.options, ModelConfig)


def test_evaluate(model_config, model_options):
    population = FeedForward(3, 2, 0.5, 3, model_options)
    population.individuals = OrderedDict()

    for i in range(3):
        individual = FeedForwardIndividual(model_config, population)
        individual.get_fitness = MagicMock(return_value=i + 1)
        population.individuals[individual] = f"{i}"

    population._evaluate()

    sorted_individuals = list(population.individuals.keys())
    assert len(sorted_individuals) == 3
    assert sorted_individuals[0].get_fitness() == 3
    assert sorted_individuals[1].get_fitness() == 2
    assert sorted_individuals[2].get_fitness() == 1


def test_kill_off_worst(model_config, model_options):
    population = FeedForward(4, 2, 0.5, 3, model_options)
    population.individuals = OrderedDict()

    for i in range(4):
        print(model_config)
        individual = FeedForwardIndividual(model_config, population)
        individual.get_fitness = MagicMock(return_value=i + 1)
        population.individuals[individual] = f"{i}"

    population._evaluate()

    population._kill_off_worst()

    remaining_individuals = list(population.individuals.keys())
    assert len(remaining_individuals) == 2
    assert remaining_individuals[0].get_fitness() == 4
    assert remaining_individuals[1].get_fitness() == 3


def test_reproduce(model_config, model_options):
    population = FeedForward(4, 2, 0.5, 3, model_options)
    population.individuals = OrderedDict()

    for i in range(4):
        individual = FeedForwardIndividual(model_config, population)
        individual.get_fitness = MagicMock(return_value=i + 1)
        population.individuals[individual] = f"{i}"

    population._evaluate()
    population._reproduce()

    new_individuals = list(population.individuals.keys())
    assert len(new_individuals) == 4
    for child_name in population.individuals.values():
        if isinstance(child_name, str):
            # New children have a fitness value of "new_child_{i}", Check that
            # they exist.
            assert "child" in child_name


def test_fit(model_config, model_options, getdata, get_test_data):
    random.seed(69)
    with unittest.mock.patch.object(FeedForward, '_initialize_individuals', lambda x: None):
        population = FeedForward(4, 2, 0.2, 5, model_options)
        population.individuals = OrderedDict()

        for i in range(6):
            individual = FeedForwardIndividual(model_config, population, )
            individual.get_fitness = MagicMock(return_value=i + 1)
            population.individuals[individual] = f"{i}"

        X_train, y_train = getdata
        X_test, y_test = get_test_data
        # Test correct dataset behaviour with test data
        with unittest.mock.patch.object(FeedForwardIndividual, 'get_fitness', lambda x: random.randint(1, 10)):
            population.fit(X_train, y_train, X_test, y_test)
            assert population.get_training_data().X[0][0] == X_train[0][0]
            assert population.get_training_data().y[0][0] == 1
            assert population.best_individual is not None

            # Test correct dataset behaviour with no test data
            population.fit(X_train, y_train)
            assert population.get_testing_data() is not None
            assert population.get_training_data().X is not None



from abc import ABC, abstractmethod


def _check_individual_inputs(parents, population) -> bool:
    """
    1. Population can not be None
    2. If parents are not None, they must be a list of individuals.
    """
    if population is None:
        raise ValueError("Population can not be None.")

    if parents is not None:
        if not isinstance(parents, tuple):
            raise ValueError("Parents must be a list of individuals.")

    if parents is None:
        return True
    return False


class Individual(ABC):
    """
    Abstract class for an individual in a population. Implements common traits between all individuals.
    An individual is a KERAS Model.
    Genes are passed down from parents, or instantiated randomly.
    Individual is a base class that implements the basic functionality of an individual.
    Child classes hold specific information for different types of models.
    """

    fitness = 0

    def __init__(self, parents_genes=None, population=None):
        self.need_new_genes = _check_individual_inputs(
            parents_genes, population)
        self.population = population
        self.parents_genes = parents_genes
        self.genes = None
        # Check inputs
        self.NUM_PARENTS = None

    @abstractmethod
    def _crossover(self, *args):
        """
        Crossover genes from parents to create new genes.
        """
        pass

    @abstractmethod
    def get_fitness(self):
        """
        Get fitness of this individual.
        """
        pass



from blacklight.engine.data import BlacklightDataset
from blacklight.engine.data.blacklight_dataset import read_data_from_file, handle_pandas_data, \
    handle_numpy_data, handle_file_data, parse_target_column, determine_input_data_type
import pandas as pd
import numpy as np
import pytest

DATA_CSV = "blacklight/engine/data/tests/data/Iris.csv"
DATA_JSON = "blacklight/engine/data/tests/data/Iris.json"
DATA_PARQUET = "blacklight/engine/data/tests/data/Iris.parquet"
DATA_EXCEL = "blacklight/engine/data/tests/data/Iris.xlsx"
DATA_DIR = "blacklight/engine/data/tests/data/"


def test_parse_target_column():
    test_data = pd.read_csv(
        DATA_CSV)
    X, y = parse_target_column(test_data)
    assert X.shape == (150, 5)
    assert y.shape == (150,)
    assert isinstance(X, np.ndarray)


def test_read_error_type():
    with pytest.raises(ValueError):
        read_data_from_file(
            DATA_CSV,
            "otherdatatype")


def test_read_data_from_file():
    data = read_data_from_file(
        DATA_CSV, "csv")
    assert data.shape == (150, 6)
    data = read_data_from_file(
        DATA_JSON, "json")
    assert data.shape == (150, 6)
    data = read_data_from_file(
        DATA_PARQUET,
        "parquet")
    assert data.shape == (150, 6)
    data = read_data_from_file(
        DATA_EXCEL, "excel")
    assert data.shape == (150, 6)


def test_handle_pandas_data():
    test_data = pd.read_csv(
        DATA_CSV)
    X, y = handle_pandas_data(test_data, None)
    assert X.shape == (150, 5)
    assert y.shape == (150,)
    assert isinstance(y, np.ndarray)
    assert isinstance(X, np.ndarray)


def test_handle_numpy_data():
    test_data = pd.read_csv(
        DATA_CSV)
    X, y = handle_numpy_data(test_data.to_numpy(), None)
    assert X.shape == (150, 5)
    assert y.shape == (150,)
    assert isinstance(y, np.ndarray)
    assert isinstance(X, np.ndarray)


def test_handle_file_data():
    X, y = handle_file_data(
        DATA_CSV)
    assert X.shape == (150, 5)
    assert y.shape == (150,)
    assert isinstance(y, np.ndarray)
    assert isinstance(X, np.ndarray)


def test_determine_input_data_type():
    X_df = pd.read_csv(DATA_CSV)
    X_numpy = pd.read_csv(
        DATA_CSV).to_numpy()
    X_file = DATA_CSV
    X_directory = DATA_DIR

    assert determine_input_data_type(X_df) == "pandas"
    assert determine_input_data_type(X_numpy) == "numpy"
    assert determine_input_data_type(X_file) == "file"
    assert determine_input_data_type(X_directory) == "directory"

    with pytest.raises(TypeError):
        determine_input_data_type(1)


def test_Blacklight_Dataset():
    X_df = pd.read_csv(DATA_CSV)
    X_numpy = pd.read_csv(
        DATA_CSV).to_numpy()
    X_file = DATA_CSV

    dataset_from_df = BlacklightDataset(X_df)
    dataset_from_numpy = BlacklightDataset(X_numpy)
    dataset_from_file = BlacklightDataset(X_file)

    assert dataset_from_df.X.shape == (150, 5)
    assert dataset_from_df.y.shape == (150, 3)
    assert isinstance(dataset_from_df.y, np.ndarray)
    assert isinstance(dataset_from_df.X, np.ndarray)

    assert dataset_from_numpy.X.shape == (150, 5)
    assert dataset_from_numpy.y.shape == (150, 3)
    assert isinstance(dataset_from_numpy.y, np.ndarray)
    assert isinstance(dataset_from_numpy.X, np.ndarray)

    assert dataset_from_file.X.shape == (150, 5)
    assert dataset_from_file.y.shape == (150, 3)
    assert isinstance(dataset_from_file.y, np.ndarray)
    assert isinstance(dataset_from_file.X, np.ndarray)

    assert dataset_from_df.__getitem__(0)[0].shape == (150, 5)
    assert dataset_from_df.__getitem__(0)[1].shape == (150, 3)



import pytest
from blacklight.engine import ModelConfig

# create a new unit test, modeled after these. if we dont pass any configs, it still has all the parameters
# pytest, end with test or start with test


full_config = {

    "problem_type": "classification",  # Specifies the type of problem, which is classification.
    "input_shape": 4,  # Defines the input shape of the data.
    "min_dense_layers": 1,  # Sets the minimum number of dense layers in the neural network
    "max_dense_layers": 8,  # Sets the maximum number of dense layers in the neural network
    "min_dense_neurons": 2,  # Sets the minimum number of neurons in each dense layer
    "max_dense_neurons": 8,  # Sets the maximum number of neurons in each dense layer.
    "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"],
    # Specifies the activation functions that can be used in the dense layers.

    "target_layer": (1, "sigmoid"),
    # Sets the target layer to the first dense layer, and the activation function for this layer to "sigmoid".

    "loss": "binary_crossentropy",
    # Sets the loss function to "binary_crossentropy" which is commonly used for binary classification tasks.

    "optimizer": "adam",
    # Sets the optimizer algorithm to "adam". This is a popular optimizer algorithm for neural networks.

    "metrics": ModelConfig.get_default_metrics(),
    # Specifies the evaluation metrics to be used during training. This line likely references a method that returns
    # default metrics for the classification task.

    "learning_rate": 0.001,
    # Sets the learning rate for the optimizer to 0.001, controlling the step size during gradient descent.

    "epochs": 1000,
    # Specifies the number of training iterations the model will undergo during training.

    "batch_size": 32,
    # Sets the number of samples in each mini-batch used for training.

    "num_classes": 3,
    # Sets the number of classes in the classification task to 3.

    "verbose": 0,
    # Sets the verbosity level during training. Default is 0 (silent).

    "class_weight": None,
    # Sets the class weights for handling imbalanced datasets. Default is None (equal weights).

    "validation_data": None,
    # Sets the validation data to be used during training. Default is None (no validation data).

    "use_multiprocessing": False,
    # Specifies whether to use multiprocessing during training. Default is False.

    "early_stopping": True,
    # Specifies whether to use early stopping during training. Default is True.

    "callbacks": ModelConfig.get_default_callbacks(),
    # Specifies the callbacks to be used during training. Default is a method that returns default callbacks.

    "output_bias": None,
    # Sets the output bias for the model. Default is None (no output bias).

    "fitness_metric": "auc"
    # Sets the fitness metric to be used for evaluating the model's performance. Default is "auc".
}

default_config = {
    "problem_type": "classification",
    "input_shape": 4,
    "min_dense_layers": 1,
    "max_dense_layers": 8,
    "min_dense_neurons": 2,
    "max_dense_neurons": 8,
    "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"],
    "target_layer": (1, "sigmoid"),
    "loss": "binary_crossentropy",
    "optimizer": "adam",
    "metrics": ModelConfig.get_default_metrics(),
    "learning_rate": 0.001,
    "epochs": 1000,
    "batch_size": 32,
    "num_classes": 3, }

partial_config = {
    "min_dense_layers": 2,
    "max_dense_layers": 91,
    "min_dense_neurons": 2,
    "max_dense_neurons": 9,
    "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"],
    "target_layer": (1, "sigmoid"),
    "loss": "ammar",
    "problem_type": "regression",
    "num_classes": 9, }


@pytest.mark.parametrize("config", [None, full_config, partial_config])
def test_model_config(config):
    model_config = ModelConfig(config)
    assert isinstance(model_config, ModelConfig)
    for key in default_config.keys():
        assert key in model_config.config


def test_other_problem_types_parse():
    config_one = {
        "loss": "binary_crossentropy",
        "optimizer": "adam",
        "metrics": ModelConfig.get_default_metrics(),
        "learning_rate": 0.001,
        "epochs": 1000,
        "batch_size": 32,
        "problem_type": "binary_classification"
    }
    config_two = {
        "loss": "mse",
        "optimizer": "adam",
        "metrics": ModelConfig.get_default_metrics(),
        "learning_rate": 0.001,
        "epochs": 1000,
        "batch_size": 32,
        "problem_type": "regression"
    }
    mc = ModelConfig(config_one)
    rc = ModelConfig(config_two)
    assert mc.config is not None
    assert rc.config is not None



from typing import Optional
from tensorflow import keras


class ModelConfig:
    def __init__(self, config: Optional[dict] = None):

        model_options = {
            "problem_type": "classification",
            "input_shape": 4,
            "min_dense_layers": 1,
            "max_dense_layers": 8,
            "min_dense_neurons": 2,
            "max_dense_neurons": 8,
            "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"],
            "target_layer": (1, "sigmoid"),
            "loss": "binary_crossentropy",
            "optimizer": "adam",
            "metrics": ModelConfig.get_default_metrics(),
            "learning_rate": 0.001,
            "epochs": 1000,
            "batch_size": 32,
            "num_classes": 3, }

        default_config = model_options

        # Update default_config with values from config (if not None)
        if config is not None:
            for key in config.keys():
                if key in default_config:
                    default_config[key] = config[key]

        # Update default_config based on the problem type
        if default_config.get("problem_type") == "classification":
            if "num_classes" not in default_config:
                raise ValueError("ModelConfig has no num_classes set for problem type Classification.")
            default_config["target_layer"] = (default_config.get("num_classes"), "softmax")
            default_config["loss"] = "categorical_crossentropy"
        elif default_config.get("problem_type") == "binary_classification":
            default_config["target_layer"] = (1, "sigmoid")
            default_config["loss"] = "binary_crossentropy"
            default_config["num_classes"] = 2
        elif default_config.get("problem_type") == "regression":
            default_config["target_layer"] = (1, "linear")
            default_config["loss"] = "mse"

        # Set default values for the remaining attributes, not set by the user
        default_config["verbose"] = default_config.get("verbose", 0)
        default_config["class_weight"] = default_config.get("class_weight", None)
        default_config["validation_data"] = default_config.get("validation_data", None)
        default_config["use_multiprocessing"] = default_config.get("use_multiprocessing", False)
        default_config["early_stopping"] = default_config.get("early_stopping", True)
        default_config["callbacks"] = default_config.get("callbacks", ModelConfig.get_default_callbacks())
        default_config["output_bias"] = default_config.get("output_bias", None)
        default_config["fitness_metric"] = default_config.get("fitness_metric", "auc")

        # Set self.config to the final default_config
        self.config = default_config
        self.check_config()

    def get(self, key: str, default=None):
        if key is None or self.config is None:
            return ValueError("Key cannot be None")
        else:
            return self.config.get(key, default)

    def check_config(self):
        print(self.config)

        if self.config is None:
            raise ValueError("ModelConfig has no config set.")
        if "target_layer" not in self.config:
            raise ValueError("ModelConfig has no target_layer set.")
        if "loss" not in self.config:
            raise ValueError("ModelConfig has no loss set.")
        if "optimizer" not in self.config:
            raise ValueError("ModelConfig has no optimizer set.")
        if "metrics" not in self.config:
            raise ValueError("ModelConfig has no metrics set.")
        if "learning_rate" not in self.config:
            raise ValueError("ModelConfig has no learning_rate set.")
        if "epochs" not in self.config:
            raise ValueError("ModelConfig has no epochs set.")
        if "batch_size" not in self.config:
            raise ValueError("ModelConfig has no batch_size set.")
        if self.config.get("problem_type") == "classification":
            if "num_classes" not in self.config:
                raise ValueError(
                    "ModelConfig has no num_classes set for problem type Classification.")
            if self.config.get("loss") != "categorical_crossentropy":
                raise ValueError(
                    f"ModelConfig has invalid loss {self.config.get('loss')} for problem type multiclass classification.")
        elif self.config.get("problem_type") == "binary_classification":
            if self.config.get("loss") != "binary_crossentropy":
                raise ValueError(
                    f"ModelConfig has invalid loss {self.config.get('loss')} for problem type binary classification.")

    @staticmethod
    def get_default_metrics():
        return [
            keras.metrics.TruePositives(name='tp'),
            keras.metrics.FalsePositives(name='fp'),
            keras.metrics.TrueNegatives(name='tn'),
            keras.metrics.FalseNegatives(name='fn'),
            keras.metrics.CategoricalAccuracy(name='accuracy'),
            keras.metrics.Precision(name='precision'),
            keras.metrics.Recall(name='recall'),
            keras.metrics.AUC(name='auc'),
        ]

    @staticmethod
    def get_default_callbacks():
        return keras.callbacks.EarlyStopping(
            monitor='auc',
            verbose=1,
            patience=10,
            mode='max',
            restore_best_weights=True)



from collections import OrderedDict
import numpy as np


class Population:
    """
    An abstract class representing a population of Neural Network topologies that can be evaluated. Their GOAL is the current dataset that this population wants to work on.
    """

    def __init__(
            self,
            number_of_individuals: int,
            num_parents_mating: int,
            death_percentage: float,
            number_of_generations: int,
            options):
        self.test_data = None
        self.num_individuals = number_of_individuals
        self.num_parents_mating = num_parents_mating
        self.num_generations = number_of_generations
        self.death_percentage = death_percentage
        self.problem_type = None
        self.num_classes = 3
        self.individuals = None
        self.train_data = None
        self.test_data = None

    def simulate(self):
        """
        Simulate the population for the number of generations.
        Callable method.
        :return:
        """
        pass

    def _evaluate(self):
        """
        Evaluate the fitness of all individuals in the population.
        :return:
        """
        if not self.individuals:
            raise ValueError("No individuals in population.")
        evaluated_individuals = OrderedDict(
            (individual.get_fitness,
             individual) for individual in self.individuals.values())
        self.individuals = OrderedDict(sorted(
            evaluated_individuals.items(),
            key=lambda x: x[0],
            reverse=True))

    def _kill_off_worst(self):
        if not self.individuals:
            raise ValueError("No individuals in population.")
        for _ in range(
                int((self.death_percentage * self.num_individuals))):
            self.individuals.popitem()
            self.num_individuals -= 1

    def _reproduce(self):
        self._kill_off_worst()
        if not self.individuals:
            raise ValueError("No individuals in population.")
        for i in range(self.num_parents_mating):
            parents = np.random.choice(
                list(self.individuals.keys()), size=2, replace=False)
            child = parents[0].mate(parents[1])
            # Add individual from Bayesian Optimization.
            self.individuals[child] = f"new_child_{i}"
            self.num_individuals += 1

    def get_training_data(self):
        return self.train_data if self.train_data else None

    def get_testing_data(self):
        return self.test_data if self.test_data else None

    def _get_fitness(self):
        pass



from blacklight.genetic.populations import FeedForward
import pandas as pd
import pytest


@pytest.mark.slow
def test_end_to_end_feed_forward_NN():
    data = pd.read_csv('blacklight/genetic/populations/tests/data/Iris.csv', index_col=0)
    model_options = {
        "layer_information": {
            "problem_type": "classification",
            "input_shape": 4,
            "min_dense_layers": 1,
            "max_dense_layers": 8,
            "min_dense_neurons": 10,
            "max_dense_neurons": 50,
            "dense_activation_types": ["relu", "sigmoid", "tanh", "selu"]
        },
        "problem_type": "classification",
        "num_classes": 3,
    }
    ff_autoML = FeedForward(8, 2, 0.1, 4, model_options)
    # Fit the model
    ff_autoML.fit(data)
    # Get model
    ff_autoML.print_model_summary()
    ff_autoML.print_model_training_history()
